@@ -4,10 +4,12 @@ import os
 import sys
 import timeit
 
-from datetime import datetime
+from datetime import datetime, timedelta
 
 import pytz
 
+if (sys.version_info.major, sys.version_info.minor) >= (3, 5):
+    from metomi.isodatetime.data import TimePoint
 
 try:
     from importlib.metadata import version as get_module_version
@@ -17,32 +19,58 @@ except ImportError:
 ISO_8601_MODULES = {
     "aniso8601": ("import aniso8601", "aniso8601.parse_datetime('{timestamp}')"),
     "ciso8601": ("import ciso8601", "ciso8601.parse_datetime('{timestamp}')"),
+    "hardcoded": ("import ciso8601", "ciso8601._hard_coded_benchmark_timestamp()"),
     "python-dateutil": ("import dateutil.parser", "dateutil.parser.parse('{timestamp}')"),
     "iso8601": ("import iso8601", "iso8601.parse_date('{timestamp}')"),
     "isodate": ("import isodate", "isodate.parse_datetime('{timestamp}')"),
-    "maya": ("import maya", "maya.parse('{timestamp}').datetime()"),
-    "pendulum": ("from pendulum.parsing import parse_iso8601", "parse_iso8601('{timestamp}')"),
     "PySO8601": ("import PySO8601", "PySO8601.parse('{timestamp}')"),
     "str2date": ("from str2date import str2date", "str2date('{timestamp}')"),
 }
 
-if os.name != "nt" and (sys.version_info.major, sys.version_info.minor) < (3, 9):
+if (sys.version_info.major, sys.version_info.minor) >= (3, 11):
+    # Python 3.11 added full ISO 8601 parsing
+    ISO_8601_MODULES["datetime (builtin)"] = ("from datetime import datetime", "datetime.fromisoformat('{timestamp}')")
+
+if sys.version_info.major >= 3 and (sys.version_info.major, sys.version_info.minor) < (3, 11):
+    # backports.datetime_fromisoformat brings the Python 3.11 logic into older Python 3 versions
+    ISO_8601_MODULES["backports.datetime_fromisoformat"] = ("from backports.datetime_fromisoformat import datetime_fromisoformat", "datetime_fromisoformat('{timestamp}')")
+
+if os.name != "nt":
     # udatetime doesn't support Windows.
     ISO_8601_MODULES["udatetime"] = ("import udatetime", "udatetime.from_string('{timestamp}')")
 
+if (sys.version_info.major, sys.version_info.minor) >= (3, 5):
+    # metomi-isodatetime doesn't support Python < 3.5
+    ISO_8601_MODULES["metomi-isodatetime"] = ("import metomi.isodatetime.parsers as parse", "parse.TimePointParser().parse('{timestamp}')")
+
 if (sys.version_info.major, sys.version_info.minor) >= (3, 6):
     # zulu v2.0.0+ no longer supports Python < 3.6
     ISO_8601_MODULES["zulu"] = ("import zulu", "zulu.parse('{timestamp}')")
 
-if (sys.version_info.major, sys.version_info.minor) != (3, 6):
+if (sys.version_info.major, sys.version_info.minor) != (3, 6) and (sys.version_info.major, sys.version_info.minor) <= (3, 9):
     # iso8601utils installs enum34, which messes with tox in Python 3.6
     # https://stackoverflow.com/q/43124775
+    # https://github.com/silverfernsys/iso8601utils/pull/5
+    # iso8601utils uses `from collections import Iterable` which no longer works in Python 3.10
+    # https://github.com/silverfernsys/iso8601utils/issues/6
     ISO_8601_MODULES["iso8601utils"] = ("from iso8601utils import parsers", "parsers.datetime('{timestamp}')")
 
 if (sys.version_info.major, sys.version_info.minor) != (3, 4):
-    # arrow no longer supports Python 3.4
+    # `arrow` no longer supports Python 3.4
     ISO_8601_MODULES["arrow"] = ("import arrow", "arrow.get('{timestamp}').datetime")
-    # moment is built on `times`, which is built on `arrow`, which no longer supports Python 3.4
+
+if sys.version_info.major >= 3 and (sys.version_info.major, sys.version_info.minor) < (3, 12):
+    # `maya` uses a version of `regex` which no longer supports Python 2
+    # `maya` uses `pendulum`, which doesn't yet support Python 3.12
+    ISO_8601_MODULES["maya"] = ("import maya", "maya.parse('{timestamp}').datetime()")
+
+if (sys.version_info.major, sys.version_info.minor) < (3, 12):
+    # `pendulum` doesn't yet support Python 3.12
+    ISO_8601_MODULES["pendulum"] = ("from pendulum.parsing import parse_iso8601", "parse_iso8601('{timestamp}')")
+
+if (sys.version_info.major, sys.version_info.minor) >= (3, 5):
+    # `moment` is built on `times`, which is built on `arrow`, which no longer supports Python 3.4
+    # `moment` uses a version of `regex` which no longer supports Python 2
     ISO_8601_MODULES["moment"] = ("import moment", "moment.date('{timestamp}').date")
 
 class Result:
@@ -68,11 +96,29 @@ class Result:
             self.exception
         ]
 
+def metomi_compare(timepoint, dt):
+    # Really (s)crappy comparison function
+    # Ignores subsecond accuracy.
+    # https://github.com/metomi/isodatetime/issues/196
+    offset = timedelta(hours=timepoint.time_zone.hours, minutes=timepoint.time_zone.minutes)
+    return timepoint.year == dt.year and \
+        timepoint.month_of_year == dt.month and \
+        timepoint.day_of_month == dt.day and \
+        timepoint.hour_of_day == dt.hour and \
+        timepoint.minute_of_hour == dt.minute and \
+        timepoint.second_of_minute == dt.second and \
+        offset == dt.tzinfo.utcoffset(dt)
+
 def check_roughly_equivalent(dt1, dt2):
     # For the purposes of our benchmarking, we don't care if the datetime
     # has tzinfo=UTC or is naive.
     dt1 = dt1.replace(tzinfo=pytz.UTC) if isinstance(dt1, datetime) and dt1.tzinfo is None else dt1
     dt2 = dt2.replace(tzinfo=pytz.UTC) if isinstance(dt2, datetime) and dt2.tzinfo is None else dt2
+
+    # Special handling for metomi-isodatetime
+    if (sys.version_info.major, sys.version_info.minor) >= (3, 5) and isinstance(dt1, TimePoint):
+        return metomi_compare(dt1, dt2)
+
     return dt1 == dt2
 
 def auto_range_counts(filepath):
@@ -105,7 +151,15 @@ def write_module_versions(filepath):
         module_version_writer = csv.writer(fout, delimiter=",", quotechar='"', lineterminator="\n")
         module_version_writer.writerow([sys.version_info.major, sys.version_info.minor])
         for module, (_setup, _stmt) in sorted(ISO_8601_MODULES.items(), key=lambda x: x[0].lower()):
-            module_version_writer.writerow([module, get_module_version(module)])
+            if module == "datetime (builtin)" or module == "hardcoded":
+                continue
+            # Unfortunately, `backports.datetime_fromisoformat` has the distribution name `backports-datetime-fromisoformat` in PyPI
+            # This messes with Python 3.8 and 3.9's get_module_version, so we special case it.
+            if module == "backports.datetime_fromisoformat":
+                module_version = get_module_version("backports-datetime-fromisoformat")
+            else:
+                module_version = get_module_version(module)
+            module_version_writer.writerow([module, module_version])
 
 def run_tests(timestamp, results_directory, compare_to):
     # `Timer.autorange` only exists in Python 3.6+. We want the tests to run in a reasonable amount of time,
