@@ -6,17 +6,14 @@ import pickle
 import platform
 import re
 import sys
+import unittest
 
-from ciso8601 import FixedOffset, parse_datetime, parse_datetime_as_naive, parse_rfc3339
+from ciso8601 import _hard_coded_benchmark_timestamp, FixedOffset, parse_datetime, parse_datetime_as_naive, parse_rfc3339
 from generate_test_timestamps import generate_valid_timestamp_and_datetime, generate_invalid_timestamp
 
 if sys.version_info.major == 2:
-    # We use unittest2 since it has a backport of the `unittest.TestCase.assertRaisesRegex` method,
-    # which is called `assertRaisesRegexp` in Python 2. This saves us the hassle of monkey-patching
-    # the class ourselves.
-    import unittest2 as unittest
-else:
-    import unittest
+    # We use add `unittest.TestCase.assertRaisesRegex` method, which is called `assertRaisesRegexp` in Python 2.
+    unittest.TestCase.assertRaisesRegex = unittest.TestCase.assertRaisesRegexp
 
 
 class ValidTimestampTestCase(unittest.TestCase):
@@ -59,36 +56,75 @@ class ValidTimestampTestCase(unittest.TestCase):
 
     def test_returns_built_in_utc_if_available(self):
         # Python 3.7 added a built-in UTC object at the C level (`PyDateTime_TimeZone_UTC`)
-        # PyPy added support for it in 7.3.6
+        # PyPy added support for it in 7.3.6, but only for PyPy 3.8+
 
         timestamp = '2018-01-01T00:00:00.00Z'
-        if (platform.python_implementation() == 'CPython' and sys.version_info >= (3, 7)) or \
-           (platform.python_implementation() == 'PyPy' and sys.pypy_version_info >= (7, 3, 6)):
+        if sys.version_info >= (3, 7) and \
+            (platform.python_implementation() == 'CPython'
+             or (platform.python_implementation() == 'PyPy' and sys.version_info >= (3, 8) and sys.pypy_version_info >= (7, 3, 6))):
             self.assertIs(parse_datetime(timestamp).tzinfo, datetime.timezone.utc)
         else:
             self.assertIsInstance(parse_datetime(timestamp).tzinfo, FixedOffset)
 
+    def test_ordinal(self):
+        self.assertEqual(
+            parse_datetime("2014-001"),
+            datetime.datetime(2014, 1, 1, 0, 0, 0),
+        )
+        self.assertEqual(
+            parse_datetime("2014-031"),
+            datetime.datetime(2014, 1, 31, 0, 0, 0),
+        )
+        self.assertEqual(
+            parse_datetime("2014-032"),
+            datetime.datetime(2014, 2, 1, 0, 0, 0),
+        )
+        self.assertEqual(
+            parse_datetime("2014-059"),
+            datetime.datetime(2014, 2, 28, 0, 0, 0),
+        )
+        self.assertEqual(
+            parse_datetime("2014-060"),
+            datetime.datetime(2014, 3, 1, 0, 0, 0),
+        )
+        self.assertEqual(
+            parse_datetime("2016-060"), # Leap year
+            datetime.datetime(2016, 2, 29, 0, 0, 0),
+        )
+        self.assertEqual(
+            parse_datetime("2014-365"),
+            datetime.datetime(2014, 12, 31, 0, 0, 0),
+        )
+        self.assertEqual(
+            parse_datetime("2016-365"), # Leap year
+            datetime.datetime(2016, 12, 30, 0, 0, 0),
+        )
+        self.assertEqual(
+            parse_datetime("2016-366"), # Leap year
+            datetime.datetime(2016, 12, 31, 0, 0, 0),
+        )
+
 class InvalidTimestampTestCase(unittest.TestCase):
     # Many invalid test cases are covered by `test_parse_auto_generated_invalid_formats`,
     # But it doesn't cover all invalid cases, so we test those here.
     # See `generate_test_timestamps.generate_invalid_timestamp` for details.
 
     def test_parse_auto_generated_invalid_formats(self):
-        for timestamp in generate_invalid_timestamp():
+        for timestamp, reason in generate_invalid_timestamp():
             try:
-                with self.assertRaises(ValueError, msg="Timestamp '{0}' was supposed to be invalid, but parsing it didn't raise ValueError.".format(timestamp)):
+                with self.assertRaises(ValueError, msg="Timestamp '{0}' was supposed to be invalid ({1}), but parsing it didn't raise ValueError.".format(timestamp, reason)):
                     parse_datetime(timestamp)
             except Exception as exc:
-                print("Timestamp '{0}' was supposed to raise ValueError, but raised {1} instead".format(timestamp, type(exc).__name__))
+                print("Timestamp '{0}' was supposed to raise ValueError ({1}), but raised {2} instead".format(timestamp, reason, type(exc).__name__))
                 raise
 
     def test_non_ascii_characters(self):
         if sys.version_info >= (3, 3):
             self.assertRaisesRegex(
                 ValueError,
-                r"Invalid character while parsing date separator \('-'\) \('🐵', Index: 7\)",
+                r"Invalid character while parsing date and time separator \(i.e., 'T', 't', or ' '\) \('🐵', Index: 10\)",
                 parse_datetime,
-                "2019-01🐵01",
+                "2019-01-01🐵01:02:03Z",
             )
             self.assertRaisesRegex(
                 ValueError,
@@ -99,7 +135,7 @@ class InvalidTimestampTestCase(unittest.TestCase):
         else:
             self.assertRaisesRegex(
                 ValueError,
-                r"Invalid character while parsing date separator \('-'\) \(Index: 7\)",
+                r"Invalid character while parsing ordinal day \(Index: 7\)",
                 parse_datetime,
                 "2019-01🐵01",
             )
@@ -120,21 +156,21 @@ class InvalidTimestampTestCase(unittest.TestCase):
 
         self.assertRaisesRegex(
             ValueError,
-            r"Invalid character while parsing date separator \('-'\) \('=', Index: 7\)",
+            r"Invalid character while parsing ordinal day \('=', Index: 7\)",
             parse_datetime,
             "2018-01=01",
         )
 
         self.assertRaisesRegex(
             ValueError,
-            r"Invalid character while parsing date separator \('-'\) \('0', Index: 7\)",
+            r"Invalid character while parsing date and time separator \(i.e., 'T', 't', or ' '\) \('2', Index: 8\)",
             parse_datetime,
-            "2018-0101",
+            "2018-0102",
         )
 
         self.assertRaisesRegex(
             ValueError,
-            r"Invalid character while parsing day \('-', Index: 6\)",
+            r"Invalid character while parsing ordinal day \('-', Index: 6\)",
             parse_datetime,
             "201801-01",
         )
@@ -263,10 +299,53 @@ class InvalidTimestampTestCase(unittest.TestCase):
                 "2014-06-00",
             )
 
+    def test_invalid_ordinal(self):
+        self.assertRaisesRegex(
+            ValueError,
+            r"Invalid ordinal day: 0 is too small",
+            parse_datetime,
+            "2014-000",
+        )
+
+        self.assertRaisesRegex(
+            ValueError,
+            r"Invalid ordinal day: 0 is too small",
+            parse_datetime,
+            "2014000",
+        )
+
+        self.assertRaisesRegex(
+            ValueError,
+            r"Invalid ordinal day: 366 is too large for year 2014",
+            parse_datetime,
+            "2014-366", # Not a leap year
+        )
+
+        self.assertRaisesRegex(
+            ValueError,
+            r"Invalid ordinal day: 366 is too large for year 2014",
+            parse_datetime,
+            "2014366", # Not a leap year
+        )
+
+        self.assertRaisesRegex(
+            ValueError,
+            r"Invalid ordinal day: 999 is too large for year 2014",
+            parse_datetime,
+            "2014-999",
+        )
+
+        self.assertRaisesRegex(
+            ValueError,
+            r"Invalid ordinal day: 999 is too large for year 2014",
+            parse_datetime,
+            "2014999",
+        )
+
     def test_invalid_yyyymm_format(self):
         self.assertRaisesRegex(
             ValueError,
-            r"Unexpected end of string while parsing day. Expected 2 more characters",
+            r"Unexpected end of string while parsing ordinal day. Expected 1 more character",
             parse_datetime,
             "201406",
         )
@@ -274,7 +353,7 @@ class InvalidTimestampTestCase(unittest.TestCase):
     def test_invalid_date_and_time_separator(self):
         self.assertRaisesRegex(
             ValueError,
-            r"Invalid character while parsing date and time separator \(ie. 'T' or ' '\) \('_', Index: 10\)",
+            r"Invalid character while parsing date and time separator \(i.e., 'T', 't', or ' '\) \('_', Index: 10\)",
             parse_datetime,
             "2018-01-01_00:00:00",
         )
@@ -505,7 +584,7 @@ class GithubIssueRegressionTestCase(unittest.TestCase):
     def test_issue_35(self):
         self.assertRaisesRegex(
             ValueError,
-            r"Invalid character while parsing date separator \('-'\) \('1', Index: 7\)",
+            r"Invalid character while parsing date and time separator \(i.e., 'T', 't', or ' '\) \('2', Index: 8\)",
             parse_datetime,
             "2017-0012-27T13:35:19+0200",
         )
@@ -542,5 +621,12 @@ class GithubIssueRegressionTestCase(unittest.TestCase):
         )
 
 
+class HardCodedBenchmarkTimestampTestCase(unittest.TestCase):
+    def test_returns_expected_hardcoded_datetime(self):
+        self.assertEqual(
+            _hard_coded_benchmark_timestamp(),
+            datetime.datetime(2014, 1, 9, 21, 48, 0, 0),
+        )
+
 if __name__ == "__main__":
     unittest.main()
